"""Fixpoint iterator over a CfG

- Transfer is a base class that must be instantiated for
each static analysis.
- fixpoint_iteration computes a fixpoint over a cfg for the given
transfer instance
"""
from cfg import *
from typing import Protocol

class Transfer[S](Protocol):
    """lists the functions that must be implemented for a static analysis over Lattice S"""
    def bottom(self) -> S:
        """the empty state, associated to unvisited nodes"""
        ...
    def init_state(self) -> S:
        """the state associated to the initial node"""
        ...
    def join(self,s1:S,s2:S) -> S:
        """join of two states when merging branches"""
        ...
    def included(self,s1: S,s2: S) -> bool:
        """return True when the first state is included in the second"""
        ...
    def tr_skip(self,s: S) -> S:
        """transfer state across a Skip transition

        Almost always transfer the state unchanged
        """
        return s
    def tr_set(self,s: S,v: str,e: ArithExpr) -> S:
        """transfer state across an assignment"""
        ...
    def tr_test(self,s: S,c: BoolExpr) -> S:
        """transfer state across a test transition"""
        ...
    def tr_err(self,s: S,e: Expr) -> S:
        """transfer state across a transition to error state"""
        ...

def fixpoint_iteration[T](transfer: Transfer[T], cfg: Cfg) -> dict[Node,T]:
    """computes the fixpoint for the given transfer functions over the given CfG"""

    env : dict[Node, T] = {}
    # map bottom to all nodes
    for node in cfg.g.nodes:
        env = env | {node : transfer.bottom()}
    # except map initial state to initial node
    env = env | {cfg.init_node : transfer.init_state()}
    # build worklist with all the nodes from the initial node
    worklist: list[tuple[Node, Node, Label]]= []
    for edge in cfg.init_node.succs:
        worklist.append((cfg.init_node,) + edge)

    # as long as the worklist is not empty
    while worklist != []:
        e = worklist.pop()
        src = e[0]
        dst = e[1]
        l   = e[2]

        s1 = env[dst]
        s_src = env[src]
        s2 = s1 # to handled elegantly unmatched cases
        match l:
            case LSet(var, expr):
                s2 = transfer.tr_set(s_src, var, expr)
            case LTest(cond):
                s2 = transfer.tr_test(s_src, cond)
            case LErr(err):
                s2 = transfer.tr_err(s_src, err)
            case LSkip():
                s2 = transfer.tr_skip(s_src)
            case _:
                # What should we do here ? Isn't this supposed to be unreachable ?
                pass

        env = env | {dst : transfer.join(s1, s2)}
        if not(transfer.included(env[dst], s1)):
            for edge in dst.succs:
                worklist.append((dst,) + edge)
        
    return env
